#include <cstddef>
using std::size_t;

#include <cstdlib>
using std::exit;

#include <cassert>

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

#include <algorithm>
using std::swap;

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <random>
using std::mt19937;
using std::uniform_real_distribution;

#include <chrono>
using std::chrono::high_resolution_clock;

#include <functional>
using std::bind;

static mt19937 rng; // mersenne twistor random number generator
static auto rnd = bind(uniform_real_distribution<double>(0,1),rng); // random number in [0,1)

const size_t L = 400;
const size_t halfL = L/2;
const size_t quartL = halfL/2;
char fluid[L][L];

const int delay = 50; // milliseconds

double pos(size_t i) { return 2.0*i/L-1.0; }

void display()
{
	glClear(GL_COLOR_BUFFER_BIT);
	for (size_t i = 0; i < L; ++i)
		for (size_t j = 0; j < L; ++j)
			if (fluid[i][j] != 0)
			{
				if (fluid[i][j] == 1)
					glColor3f(1.0, 1.0, 1.0); 
				else
					glColor3f(1.0, 0.5, 0.25); 
				glBegin(GL_POLYGON);
					glVertex2f(pos(i),pos(j));
					glVertex2f(pos(i+1),pos(j));
					glVertex2f(pos(i+1),pos(j+1));
					glVertex2f(pos(i),pos(j+1));
				glEnd();
			}
	glutSwapBuffers();
}

void keyboard(unsigned char key, int x, int y)
{
	if (key == 'q')
		exit(0);
}

int setup;
double concentration;
bool check_conservation_laws = true;
void initialize_grid(void)
{
	switch(setup)
	{
	case 0: // fluid 
		for (size_t i = 0; i < L; ++i)
			for (size_t j = 0; j < L; ++j)
				fluid[i][j] = (rnd() < concentration ? 1 : 0);
		break;

	case 1: // fluid with a low density hole
		check_conservation_laws=false;
		for (size_t i = 0; i < L; ++i)
			for (size_t j = 0; j < L; ++j)
			{
				const double x = double(i) - double(quartL);
				const double y = double(j) - double(quartL);
				const double c = (x*x + y*y < 0.5*quartL*quartL ? 0.1*concentration : concentration);
				fluid[i][j] = (rnd() < c ? 1 : 0);
			}
		break;
		
	case 2: // fluid in a container
		check_conservation_laws=false;
		for (size_t i = 3; i < halfL; ++i)
			for (size_t j = 3; j < L-3; ++j)
				fluid[i][j] = (rnd() < concentration ? 1 : 0);
		
			for (size_t i = 0; i < L; ++i)
		{
			fluid[i][0] = 0;
			fluid[i][L-1] = 0;
			fluid[0][i] = 0;
			fluid[L-1][i] = 0;
		}

		for (size_t i = 1; i < L-1; ++i)
		{
			fluid[i][1] = fluid[i][2] = 2;
			fluid[i][L-2] = fluid[i][L-3] = 2;
			fluid[1][i] = fluid[2][i] = 2;
			fluid[L-2][i] = fluid[L-3][i] = 2;
			if (i < L/3-2 or i > L/3+2)
				fluid[halfL][i] = fluid[halfL+1][i] = 2;
		}

		break;

	default:
		assert(false);
		break;
	}
}

void update(char &TL, char &TR, char &BL, char &BR)
{
	/*
	// local momentum px = TL-BR; py = TR-BL; collision when vx == vy == 0
	//
	// zero particles:
	// 00 -> 00     
	// 00    00     
	//
	// one particle:
    // 10 -> 00     01 -> 00     00 -> 10     00 -> 01
	// 00    01     00    10     01    00     10    00
	//
	// two particles:
	// 10 -> 01     01 -> 10     10 -> 01     11 -> 00     01 -> 10     00 -> 11
	// 01    10     10    01     10    01     00    11     01    10     11    00     
	//
	// three particles:
	// 01 -> 11     10 -> 11     11 -> 01     11 -> 10
	// 11    10     11    01     10    11     01    11 
	//
	// four particles
	// 11 -> 11
	// 11    11
	*/
	
	/*
	char occupancy = TL+TR+BL+BR;
    if (occupancy == 1)
	{
		if (TL == 1) { TL = 0; BR = 1; }
		else if (TR == 1) { TR = 0; BL = 1; }
		else if (BR == 1) { BR = 0; TL = 1; }
		else { assert(BL == 1); BL = 0; TR = 1; }
	}
	else if (occupancy == 3)
	{
		if (TL == 0) { TL = 1; BR = 0; }
		else if (TR == 0) { TR = 1; BL = 0; }
		else if (BR == 0) { BR = 1; TL = 0; }
		else { assert(BL == 0); BL = 1; TR = 0; }			
	}
	else if (occupancy == 2)
	{
		if (TL == 1 and BR == 1) { TL = BR = 0; BL = TR = 1; }
		else if (TR == 1 and BL == 1) { TR = BL = 0; BR = TL = 1; }
		else if (TL == 1 and BL == 1) { TL = BL = 0; TR = BR = 1; }
		else if (TL == 1 and TR == 1) { TL = TR = 0; BL = BR = 1; }
		else if (TR == 1 and BR == 1) { TR = BR = 0; TL = BL = 1; }
		else { assert(BL == 1 and BR == 1); BL = BR = 0; TL = TR = 1; }
	}
	 */

	char occupancy = TL+TR+BL+BR;
    if (occupancy%2==1)
	{
		swap(TL,BR);
		swap(TR,BL);
	}
	else if (occupancy == 2)
	{
		if (TL == BL) 
		{ 
			swap(TL,TR); 
			swap(BL,BR);
		}
		else
		{ 
			swap(TL,BL); 
			swap(TR,BR);
		}
	}	
}	

void verify(void)
{
	/* (!) write code */
	return;
}

size_t simulation_clock = 0;
void sweep_grid(void)
{
	const size_t offset = simulation_clock%2;
	for (size_t i = offset; i < L; i += 2)
		for (size_t j = offset; j < L; j += 2)
			update(fluid[i][j],fluid[(i+1)%L][j],fluid[i][(j+1)%L],fluid[(i+1)%L][(j+1)%L]);

	++simulation_clock;
	if (check_conservation_laws and simulation_clock%2==0) verify();
}
		
void animate(int)
{
	sweep_grid();
	glutPostRedisplay();
	glutTimerFunc(delay,animate,0);
}

int main(int argc, char **argv)
{
	if (argc != 3)
	{
		cerr << "Usage: hpp (setup=0,1,2) (0 < concentration < 1)" << endl;
		exit(1);
	}
	setup = atoi(argv[1]);
	concentration = atof(argv[2]);
	if (setup < 0 or setup > 2) 
	{
		cerr << "Error: `setup` parameter must be 0, 1, or 2" << endl;
		exit(0);
	}
	if (concentration < 0.0 or concentration > 1.0) 
	{
		cerr << "Error: `concentration` parameter must be in the interval [0,1]" << endl;
		exit(0);
	}
    rng.seed(mt19937::result_type(high_resolution_clock::now().time_since_epoch().count()));
	initialize_grid();	

	glutInit(&argc,argv);
	glutInitDisplayMode (GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize (800, 800);
	glutInitWindowPosition (100, 100);
	glutCreateWindow("The HPP model lattice gas");
	glutDisplayFunc(display);
    glutKeyboardFunc(keyboard);	
	glutTimerFunc(delay,animate,0);
	glutMainLoop();
}

